#! /usr/bin/env python
import argparse
import numpy as np
import pandas as pd
import pyranges as pr
import math
from scipy.stats import binom, dirichlet
from scipy.optimize import minimize, nnls, Bounds
import os
import sys
from multiprocessing import Queue
import threading
from functools import partial

script_dir = os.path.dirname(os.path.realpath(__file__))
sys.path.append(script_dir)

from _nanomix import *
from atlas import AtlasMethylome
from tools import *

def fit_model(methylome, atlas, model, p01, p11,
              threads=1, n_trials=5,
              nnls_init=False, concentration=1.0):
    """
    Wrapper function to select model for deconvolution
    mmse model is initalized seperately with Rust bindings
    methylome and data are joined on covered regions before deconvolution

    :param methylome: path to simulated methylome
    :param data_path: path to reference data
    :param model: model to fit
    :param p01: nanopore miscall rate
    :param p11: nanopore correct call rate
    :return: cell-type proportions
    """
    data = AtlasMethylome(methylome, atlas, threads=threads)
    if model == 'nnls':
        sigma = fit_nnls(data)
    elif model == 'llse':
        sigma = fit_llse_parallel(data, p01, p11, n_trials, threads, concentration, nnls_init)
    elif model == 'llsp':
        sigma = fit_llsp(data)
    elif model == 'null':
        sigma = fit_uniform(data)
    else:
        raise ValueError(f"no such model: {model}. Choose from [nnls, llse, llsp, mmse]")
    return sigma

def log_likelihood_sequencing_perfect(data, sigma):
    """
    Compute the log-likelihood of the llsp model

    :param data: reference data
    :param sigma: cell-type proportions
    :param data: data methylome
    :return: log-likelihood
    :rtype: float
    """
    sigma_t = sigma.reshape( (data.K, 1) )
    x = np.clip(np.ravel(data.get_x(sigma_t)), 0, 1.0)
    b =  binom.logpmf(data.m, data.t, x)
    return np.sum(b)

def fit_llsp(data, n_trials=10):
    """
    fit the log-likelihood sequencing perfect model

    :param data: reference data
    :param data: data methylome
    :param n_trials: number of trials to run with random initalizations of sigma
    :return: cell-type proportions
    :rtype: np.array
    """
    f = lambda x: -1 * log_likelihood_sequencing_perfect(data, x)
    bnds = [ (0.0, 1.0) ] * data.K
    cons = ({'type': 'eq', 'fun': eq_constraint})
    alpha = np.array([ 1.0 / data.K ] * data.K)
    n_trials = 10
    best_ll = np.inf
    best_sol = None
    initializations = dirichlet.rvs(alpha, size=n_trials).tolist()

    for (i, init) in enumerate(initializations):
        res = minimize(f, init, method='SLSQP', options={'maxiter': 100, 'disp':False}, bounds=bnds, constraints=cons)
        ll = res.get("fun")
        if ll < best_ll:
            best_ll = ll
            best_sol = res
    return best_sol.x/np.sum(best_sol.x)

def log_likelihood_sequencing_with_errors(data, sigma, p01, p11):
    """
    Compute the log-likelihood of the llse model

    :param data: reference data
    :param sigma: cell-type proportions
    :param data: data methylome
    :param p01: nanopore miscall rate
    :param p11: nanopore correct call rate
    :return: log-likelihood
    :rtype: float
    """

    sigma_t = sigma.reshape( (data.K, 1) )

    # if p01 is a vector then reshape if not then make it a vector
    if isinstance(p01, np.ndarray):
        p01 = p01.reshape( (data.K, 1) )
    else:
        p01 = np.array([p01] * data.K).reshape( (data.K, 1) )
    if isinstance(p11, np.ndarray):
        p11 = p11.reshape( (data.K, 1) )
    else:
        p11 = np.array([p11] * data.K).reshape( (data.K, 1) )
    p = np.clip(np.ravel(data.get_x(sigma_t, p01, p11)), 0, 1.0)
    # breakpoint()
    # x = np.clip(np.ravel(np.dot(data.A, sigma_t)), 0, 1.0)
    # p = x * p11 + (1-x) * p01
    b =  binom.logpmf(data.m, data.t, p)
    binomial_coef = sum([math.log(math.comb(int(t), int(m))) for m,t in zip(data.m, data.t)])

    return np.sum(b) #- binomial_coef

def fit_uniform(K):
    """
    fit the uniform model (null model)

    :param K: number of cell-types
    :return: cell-type proportions
    :rtype: np.array
    """
    return np.array([1.0 / K ] * K)

class my_thread(threading.Thread):
    """
    thread class for parallel processing of fit_llse
    """
    def __init__(self, data, p01, p11, queue):
        super(my_thread, self).__init__()
        self.data = data
        self.p01 = p01
        self.p11 = p11
        self.queue = queue
        self.res = []
    def run(self):
        alpha = self.queue.get()
        f = lambda x: -1 * log_likelihood_sequencing_with_errors(self.data, x, self.p01, self.p11)
        bnds = [ (0.0, 1.0) ] * self.data.K
        cons = ({'type': 'eq', 'fun': eq_constraint})
        self.res.append(minimize(f, alpha, method='SLSQP', options={'maxiter': 200, 'disp':False}, bounds=bnds, constraints=cons))


def fit_llse_parallel(data, p01, p11, n_trials, threads, concentration, init_nnls):
    alpha = np.array([concentration] * data.K)
    initializations = dirichlet.rvs(alpha, size=n_trials).tolist()
    if init_nnls:
        initializations.append(fit_nnls(data))

    queues = [Queue() for i in range(threads)]
    res = []
    for i in range(n_trials):
        queues[i % threads].put(initializations[i])
    threads = [my_thread(data, p01, p11, queues[i]) for i in range(threads)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()
        res += t.res

    best_ll = np.inf
    best_sol = None
    for r in res:
        ll = r.get("fun")
        if ll < best_ll:
            best_ll = ll
            best_sol = r
    return best_sol.x/np.sum(best_sol.x)

def fit_llse(data, p01, p11):
    """
    fit the log-likelihood sequencing with errors model

    :param data: reference data
    :param data: data methylome
    :param p01: nanopore miscall rate
    :param p11: nanopore correct call rate
    :param n_trials: number of trials to run with random initalizations of sigma
    :return: cell-type proportions
    :rtype: np.array
    """
    f = lambda x: -1 * log_likelihood_sequencing_with_errors(data, x, p01, p11)
    bnds = [ (0.0, 1.0) ] * data.K
    cons = ({'type': 'eq', 'fun': eq_constraint})
    alpha = np.array([ 1.0 / data.K ] * data.K)
    res = minimize(f, alpha, method='SLSQP', options={'maxiter': 100, 'disp':False}, bounds=bnds, constraints=cons)
    return res.x/np.sum(res.x)

def fit_nnls(data):
    """
    fit the non-negative least squares model

    :param data: reference data
    :param data: data methylome
    :return: cell-type proportions
    :rtype: np.array
    """

    # add sum=1 constraint
    t = np.array([1.0] * data.K).reshape( (1, data.K) )
    A = np.append(data.A, t, axis=0)
    b = np.append(data.x_hat, [1.0], axis=0)
    res = nnls(A, b)
    return res[0]/np.sum(res[0])

def fit_mmse(methylome, atlas, sigma, p01, p11, stop_thresh, max_iter, min_proportion, concentration,
             true_sigma=None, true_assignments=None):
    """
    Fit mixture model with sequencing errors (MMSE) onto data with reference data.
    Wrapper function to access Rust implementation.

    :param methylome: path to sample methylome to fit
    :param atlas: path to reference atlas
    :param p01: sequencing miscall rate
    :param p11: sequencing correct call rate
    :param stop_thresh: threshold for stopping iterations
    :param max_iter: maximum number of iterations
    :param min_proportion: minimum proportion of a cell type in the mixture
    :return: fitted mixture model
    """
    # initialize MMSE model
    mmse = MMSE(methylome, atlas, sigma, p01, p11, concentration)
    if true_sigma is not None and true_assignments is not None:
        mmse.evaluate(stop_thresh, max_iter, min_proportion, true_sigma, true_assignments)
    else:
        mmse.optimize(stop_thresh, max_iter, min_proportion)
    return mmse.cell_type_proportions()




import numpy as np
import csv
from collections import Counter
import os
import sys

script_dir = os.path.dirname(os.path.realpath(__file__))
sys.path.append(script_dir)

from _nanomix import *
from models import fit_model, fit_mmse, log_likelihood_sequencing_with_errors, log_likelihood_sequencing_perfect
from atlas import AtlasMethylome
from tools import *
from plot import *

def plot(sigma, outpath, chart, group_lung=False, cell_types=[]):
    """
    Wrapper function to choose plotting function

    :param sigma:
    :param outpath:
    """
    if chart == 'bar':
        plot_mixture_proportions(sigma, outpath, group_lung=group_lung)
    elif chart == 'scatter':
        plot_scatter(sigma, outpath, cell_types=cell_types)

def simulate(atlas, sigma, coverage, region_size, p01, p11):
    """
    Simulate a methylome from a reference atlas at the given cell type proportion
    Wrapper function to call Rust code from Python

    :param atlas: path to reference atlas
    :param sigma: path to tsv file containing cell type proportions
    :param coverage: coverage of simulated reads
    :param region_size: number of CpGs in each region
    :param p01: nanopore miscall rate
    :param p11: nanopore correct call rate
    :return: None
    """
    sigma = get_sigma_init(sigma, get_cell_types(atlas))
    generate_methylome(atlas, sigma, coverage, region_size, p01, p11)

def evaluate(methylome, atlas, sigma,
                p01=1.,
                p11=0.,
                sigma_init='null',
                max_iter=10,
                min_proportion=0.01,
                stop_thresh=1e-3):
    """
    Evaluate the performance of a model on a simulated methylome
    Wrapper function to call Rust code from Python

    :param methylome: path to simulated methylome
    :param atlas: path to reference atlas
    :param model: model to evaluate
    :param p01: nanopore miscall rate
    :param p11: nanopore correct call rate
    :param sigma_init: model to initialize sigma (mmse only)
    :param max_iter: maximum number of iterations (mmse only)
    :param min_proportion: minimum proportion of a cell type (mmse only)
    :param stop_thresh: stopping threshold (mmse only)
    :return: None
    """
    # Determine cell_type proportions from methylome
    cell_types = get_cell_types(atlas)
    cell_type_fragment_counts = Counter(cell_types)
    true_assignments = []

    # Open methylome and count 
    with open(methylome, 'r') as f:
        reader = csv.reader(f, delimiter='\t')
        header = next(reader)
        cell_type_idx = header.index('cell_type')
        for row in reader:
            cell_type_fragment_counts[row[cell_type_idx]] += 1
            true_assignments.append(row[cell_type_idx])


    # Convert counts to proportions
    total_fragments = sum(cell_type_fragment_counts.values())
    true_cell_type_proportions = {cell_type : count / total_fragments for cell_type, count in cell_type_fragment_counts.items()}
    true_sigma = np.array([true_cell_type_proportions[cell_type] for cell_type in cell_types])

    # open sigma.tsv file and load it
    sigma = get_sigma_init(sigma, get_cell_types(atlas))
    deconvolution_loss = np.linalg.norm(true_sigma - sigma)
    print(f"Deconvolution loss:\t{deconvolution_loss}")

    # Assign cell types
    model = MMSE(methylome, atlas, sigma, p01=p01, p11=p11, concentration=1/len(cell_types))
    cell_types += ['unassigned']
    for threshold in [.5, .6, .7, .8, .9]:
        enumerated_assignments = model.assign_fragments_t(threshold)
        assignments = [cell_types[assignment] for assignment in enumerated_assignments]
        correct = 0
        for true_assignment, assignment in zip(true_assignments, assignments):
            if 'PBMC' in true_assignment and assignment in 'monocyte erythrocyte_progenitor T-cell B-cell NK-cell CD4T-cell CD8T-cell':
                correct += 1
            elif 'CliveOME' in true_assignment and assignment in 'monocyte erythrocyte_progenitor T-cell B-cell NK-cell granulocyte CD4T-cell CD8T-cell':
                correct += 1
            elif true_assignment == assignment:
                correct += 1
        accuracy = correct / len(assignments)
        print(f"Accuracy at confidence {threshold}:\t{accuracy}")

def transform_atlas(atlas, p01, p11, threads=1):
    """
    Transform an atlas to include error parameters fixed in.
    Here p01, p11 are defined for every region in the atlas and every cell_type
    such that atlas, p01, p11 are paths to tsv files with the same column and rows

    :param atlas: path to tsv of atlas file
    :param p01: path to tsv of p01 file
    :param p11: path to tsv of p11 file
    """
    cell_types = get_cell_types(atlas)
    # columns
    columns={'chromosome':'Chromosome', 'chr':'Chromosome',
                            'start':'Start',
                            'end':'End',
                            'start_position':'Start',
                            'end_position':'End'}
    # load atlas, p01, p11 into pyranges
    atlas_pr = pd.read_csv(atlas, sep='\t').rename(columns=columns)
    p01_pr = pr.PyRanges(pd.read_csv(p01, sep='\t').rename(columns=columns))
    p11_pr = pr.PyRanges(pd.read_csv(p11, sep='\t').rename(columns=columns))

    # join p01, p11 to atlas
    pr = atlas_pr\
        .join(p01_pr, suffix=_p01, nb_cpu=threads)\
        .join(p11_pr, suffix=_p11, nb_cpu=threads)\
        .sort(nb_cpu=threads)

    # add columns for each cell type if they do not exist
    # i.e. if lung_p01 not in columns duplicate default_p01 and call it lung_p01
    for cell_type in cell_types:
        if cell_type + '_p01' not in pr.columns:
            pr[cell_type + '_p01'] = pr['default_p01']
        if cell_type + '_p11' not in pr.columns:
            pr[cell_type + '_p11'] = pr['default_p11']

    # make a numpy matrix of the original elements from atlas_pr
    A = pr.as_df()[cell_types].to_numpy()
    P01 = pr.as_df()[[f'{cell_type}{_p01}' for cell_type in cell_types]].to_numpy()
    P11 = pr.as_df()[[f'{cell_type}{_p11}' for cell_type in cell_types]].to_numpy()

    # Do element wise multiplication of A and P11
    # Do element wise multiplication of (1-A) and P01
    # Add the two matrices together
    transformed_atlas = A * P11 + (1 - A) * P01

    # write to stdout with locus and cell_type columns
    print('chromosome\tstart\tend\t' + '\t'.join(cell_types))
    for i, row in enumerate(transformed_atlas):
        row_as_string = '\t'.join([str(x) for x in row])
        print(f"{pr.as_df().iloc[i]['Chromosome']}\t{pr.as_df().iloc[i]['Start']}\t{pr.as_df().iloc[i]['End']}\t{row_as_string}")



def log_likelihood(methylome, atlas, sigma, p01, p11, model):
    """
    Return the log likelihood for a given solution

    :param methylome: Path to tsv file of methylome
    :param atlas: Path to tsv file of atlas
    :param model: Deconvolution model options: [nnls, llse, llsp, mmse]
    :param p01: Sequencing miscall rate
    :param p11: Sequencing correct call rate
    :param sigma: path to sigma file
    :return: None
    """
    atlas_path = atlas
    sigma = get_sigma_init(sigma, get_cell_types(atlas))
    ll = 'none'
    if model == 'llse':
        data = AtlasMethylome(atlas_path, methylome)
        # print("Total number of methylated calls: {}".format(np.sum(s.m)))
        # print("Total number of unmethylated calls: {}".format(np.sum(s.t-s.m)))
        # print("atlas shape: {}".format(atlas.A.shape))
        ll = log_likelihood_sequencing_with_errors(data, sigma, p01, p11)
    elif model == 'mmse':
        mmse = MMSE(methylome, atlas_path, sigma, p01, p11, 0.1)
        ll = mmse.log_likelihood()
    elif model == 'llsp':
        data = AtlasMethylome(atlas_path, methylome)
        ll = log_likelihood_sequencing_perfect(data, sigma, p01, p11)
    print("name\tvalue")
    print("Atlas\t{}".format(atlas_path.split('/')[-1].replace('.tsv', '')))
    print("Model\t{}".format(model))
    print("p01\t{}".format(p01))
    print("p11\t{}".format(p11))
    print("Methylome\t{}".format(methylome.split('/')[-1].replace('.tsv', '')))
    print("Log-likelihood\t{}".format(ll))


def deconvolute(methylome, atlas, model,
                p01=0.,
                p11=1.,
                n_trials=10,
                threads=1,
                concentration=None,
                nnls_init=False,
                sigma_init='null',
                max_iter=10,
                min_proportion=0.01,
                stop_thresh=1e-3,
                print_output=True):
    """
    Deconvolute a methylome using a given model to get the proportion of each cell type present

    :param methylome: Path to tsv file of methylome
    :param atlas: Path to tsv file of atlas
    :param model: Deconvolution model options: [nnls, llse, llsp, mmse]
    :param p01: Sequencing miscall rate
    :param p11: Sequencing correct call rate
    :param sigma_init: Initalize sigma with one of the other models (mmse only)
    :param max_iter: Maximum number of iterations for the model (mmse only)
    :param min_proportion: Minimum proportion of a cell type to be considered (mmse only)
    :param stop_thresh: Threshold for stopping iterations (mmse only)
    :return: none
    """

    # if p01 is a string, then it is filepath
    # first try and parse it into a float and then if it doesnt work then it is a filepath
    cell_types = get_cell_types(atlas)
    p01 = get_vectorized_error_param(p01, cell_types)
    p11 = get_vectorized_error_param(p11, cell_types)



    if concentration is None:
        concentration = 1/len(get_cell_types(atlas))
    # Run
    if model == 'mmse':
        if nnls_init:
            sigma = fit_model(methylome, atlas, 'nnls', p01, p11, concentration=concentration, threads=threads)
        else:
            sigma = get_sigma_init(sigma_init, get_cell_types(atlas), concentration=concentration)
        cell_type_proportions = fit_mmse(methylome, atlas, sigma, p01, p11, stop_thresh, max_iter, min_proportion, concentration)
    else:
        sigma = fit_model(methylome, atlas, model, p01, p11,
                          n_trials=n_trials,
                          threads=threads,
                          concentration=concentration,
                          nnls_init=nnls_init)
        if min_proportion > 0.0:
            sigma[sigma < min_proportion] = 0.0
            sigma /= np.sum(sigma)
        cell_type_proportions = {cell_type: proportion for cell_type, proportion in zip(get_cell_types(atlas), sigma)}

    # return deconvolution results
    if print_output:
        print("cell_type\tproportion")
        for cell_type, proportion in cell_type_proportions.items():
            print(f"{cell_type}\t{proportion}")

    return cell_type_proportions

def assign_fragments(methylome, atlas, sigma,
                     p01=0., p11=1.,
                     threshold=0.5,
                     print_output=True):
    """
    Assign fragments in the methylome to cell types
    Initialize MMSE model to use maximum responsibility (gamma) assignments

    :param atlas: path to reference atlas
    :param methylome: path to methylome
    :param sigma: path to tsv file containing cell type proportions
    :param threshold: confidence threshold for assigning fragments to cell types
    :param print_output: print output to stdout
    :return: assignments vector in order of fragments in the methylome
    """
    # Need to ensure that the cell types are in the same order as the atlas
    cell_types = get_cell_types(atlas)

    # get sigma vector
    with open(sigma, 'r') as f:
        reader = csv.reader(f, delimiter='\t')
        header = next(reader)
        cell_type_proportions = {row[0] : float(row[1]) for row in reader}

    # get sigma vector in the same order as the atlas
    sigma = np.array([cell_type_proportions[cell_type] for cell_type in cell_types])

    model = MMSE(methylome, atlas, sigma, p01=0., p11=1.)
    enumerated_assignments = model.assign_fragments_t(threshold)

    # map enumerated assignments to a vector of (string) cell type assignments
    cell_types += ['unassigned']
    cell_type_assignments = [cell_types[i] for i in enumerated_assignments]

    # return cell type assignments to standard output
    if print_output:
        with open(methylome, 'r') as f:
            reader = csv.reader(f, delimiter='\t')
            header = next(reader)
            print('\t'.join(header + ['cell_type']))
            for row, cell_type in zip(reader, cell_type_assignments):
                print('\t'.join(row + [cell_type]))

    return cell_type_assignments





import pyranges as pr
import pandas as pd
import numpy as np
from tools import *

class AtlasMethylome:
    """
    Reference atlas class for storing the methylation propensities of each cell type
    """
    def __init__(self, methylome, atlas, threads=1):
        """
        Take a methylome of read frequencies and turn it into a methylome of region frequencies from regions in the atlas

        Methylome has columns:
        read_name       chromosome      start_position  end_position    alignment_length        strand  mapping_quality total_calls     modified_calls  modification_frequency  cell_type
    e5349968-9ed4-43e4-882b-061ecc7e2683    chr1    1071849 1072048 52806   -       60      16      16      1.00    HCT116
    e5349968-9ed4-43e4-882b-061ecc7e2683    chr1    1066168 1066385 52806   -       60      6       6       1.00    HCT116


        We want to aggregate the columns modified_calls and total_calls based on regions in the first three columns of the atlas file

        :param atlas: Atlas file path
        :param methylome: Methylome file path
        :return: self
        """

        # Rename columns 
        columns={'chromosome':'Chromosome', 'chr':'Chromosome',
                                'start':'Start',
                                'end':'End',
                                'start_position':'Start',
                                'end_position':'End'}

        # Read atlas
        df_atlas = pd.read_csv(atlas, sep='\t').rename(columns=columns)
        df_atlas.drop_duplicates(inplace=True)
        if 'label' in df_atlas.columns: df_atlas.drop('label', axis=1, inplace=True)
        df_atlas.dropna(inplace=True)
        gr_atlas = pr.PyRanges(df_atlas).sort()

        # Read methylomes data from mbtools
        try:
            df = pd.read_csv(methylome, sep='\t').rename(columns=columns)
        except pd.errors.EmptyDataError:
            Exception("Empty methylome file")
        df.dropna(inplace=True)
        gr_sample = pr.PyRanges(df).sort()
        gr = gr_atlas.join(gr_sample, nb_cpu=threads)

        # Check for empty upon join
        cell_types = get_cell_types(atlas)
        if len(gr) == 0:
            Exception("Empty join between atlas and sample. The sample does not overlap with any regions in the atlas.")
        df = gr.df
        df_grouped = df.groupby(['Chromosome', 'Start', 'End'], observed=True)
        df_grouped_sample = df_grouped[['modified_calls', 'total_calls']].sum().sort_index()
        df_grouped_atlas = df_grouped[cell_types].first().sort_index()
        df_join = df_grouped_sample.join(df_grouped_atlas)

        self.t = np.array(df_join.total_calls, dtype=np.float32)
        self.m = np.array(df_join.modified_calls, dtype=np.float32)
        self.x_hat = self.m/self.t

        self.cpg_ids = [(chrom, start, end) for chrom, start, end in df_join.index]
        self.K = len(cell_types)
        self.v = {k:list(df_join[k]) for k in cell_types}
        self.A = np.array(df_join.loc[:, list(cell_types)])

    def get_x(self, sigma):
        """
        Compute the expected methylome by matrix multiplication of the reference atlas and the cell-type proportions

        :param sigma: cell-type proportions
        :return: expected methylome
        """
        return np.dot(self.A, sigma)

    def get_x(self, sigma, p01, p11):
        """
        Compute the expected methylome by matrix multiplication of the reference atlas and the cell-type proportions
        with vectorized p01, p11

        :param sigma: cell-type proportions
        :param p01: p01 vector
        :param p11: p11 vector
        :return: expected methylome
        """
        return np.dot(self.A, sigma*p11) + np.dot(1-self.A, sigma*p01)

    def get_cell_types(self):
        return list(self.v.keys())

    def get_num_cell_types(self):
        return len(self.v.keys())

    def __len__(self):
        return len(self.cpg_ids)

    def __repr__(self):
        return "AtlasMethylome with {} CpGs and {} cell types".format(len(self), self.get_num_cell_types())




import csv
import numpy as np
import pandas as pd
import pyranges as pr


def get_vectorized_error_param(file, cell_types):
    """
    # we expect a tsv file with first column cell_type and the second column p01
    # open the file and read p01 values into a vector, ordered by cell_types in the atlas
    # order them by cell_types in the atlas
    """
    try:
        error_param = float(file)
        return np.array([error_param]*len(cell_types))
    except ValueError:
        with open(file, 'r') as f:
            # read tsv into a dictionary mapping cell_type to error_param
            error_param_dict = {row[0]: float(row[1]) for row in csv.reader(f, delimiter='\t')}
        error_param = []
        for cell_type in cell_types:
            if cell_type in error_param_dict:
                error_param.append(error_param_dict[cell_type])
            else:
                error_param.append(error_param_dict['default'])
        return np.array(error_param)

def eq_constraint(x):
    return 1 - np.sum(x)

def get_cell_types(atlas):
    """
    Open atlas file and read the header to get the cell types order

    :param atlas: path to reference atlas
    :return: vector of cell types in same order as the atlas
    """
    with open(atlas, 'r') as f:
        reader = csv.reader(f, delimiter='\t')
        header = next(reader)
        cell_types = header[3:]
    return cell_types

def get_sigma_init(sigma, cell_types, concentration=1.):
    """
    Get sigma vector from sigma_init file in the same order as the atlas

    :param sigma: path to sigma_init file
    :param cell_types: vector of cell types in same order as the atlas
    :return: vector of cell-type proportions
    """
    if sigma == 'null':
        return np.array([concentration] * len(cell_types))
    else:
        with open(sigma, 'r') as f:
            reader = csv.reader(f, delimiter='\t')
            header = next(reader)
            cell_type_proportions = {row[0]: float(row[1]) for row in reader}
        sigma_init = np.array([cell_type_proportions[cell_type] for cell_type in cell_types])
        return sigma_init












# Test functions in main.py
import unittest
import os
import sys


script_dir = os.path.dirname(os.path.realpath(__file__))
parent_dir = os.path.dirname(script_dir)
sys.path.append(parent_dir)

from main import *

script_dir = os.path.dirname(os.path.realpath(__file__))
methylome = os.path.join(script_dir, 'test_data', 'test_methylome.tsv')
atlas = os.path.join(script_dir, 'test_data', 'test_atlas.tsv')
sigma_path = os.path.join(script_dir, 'test_data', 'test_sigma.tsv')

class TestMain(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Write sigma
        sigma = [0.3, 0.7]
        coverage = 100
        with open(sigma_path, 'w') as f:
            f.write('cell_type\tproportion\n')
            for i, s in enumerate(sigma):
                f.write(f'cell{i}\t{s}\n')
        # Write an atlas
        with open(atlas, 'w') as f:
            f.write('chr\tstart\tend')
            for i in range(len(sigma)):
                f.write(f'\tcell{i}')
            f.write('\n')
            for i in range(len(sigma)):
                f.write(f'chr1\t{i*100}\t{i*100+99}')
                for j in range(len(sigma)):
                    if i == j:
                        f.write('\t0.99')
                    else:
                        f.write('\t0.01')
                f.write('\n')
        # Simulate a methylome
        with open(methylome, 'w') as f:
            f.write('chr\tstart\tend\ttotal_calls\tmodified_calls\n')
            for i in range(len(sigma)):
                # simulate n reads for every region
                for _ in range(coverage):
                    # choose a cell type wighted by sigma
                    cell_type = np.random.choice(len(sigma), p=sigma)
                    if cell_type == i:
                        f.write(f'chr1\t{i*100}\t{i*100+99}\t1\t1\n')
                    else:
                        f.write(f'chr1\t{i*100}\t{i*100+99}\t1\t0\n')

    def test_simulate(self):
        # Test simulate function
        simulate(atlas, sigma_path, 200, 5, 0.0, 1.0)

        # check that the cell_type proportions are close to sigma
        # Tissues are enumerated as 0 and 1. so the average of them should be the proportion of tissue2
        # with open(methylome, 'r') as f:
            # lines = f.readlines()
            # cell_types = [int(line.split('\t')[-1].strip()) for line in lines[1:]]
        # self.assertAlmostEqual(sum(cell_types)/len(cell_types), 0.7, places=1)

if __name__ == '__main__':
    unittest.main()



#! /usr/bin/env python
#############################################################
#                                                           #
#  Code adapted from https://github.com/nloyfer/meth_atlas  #
#                                                           #
#############################################################
import numpy as np
import pandas as pd
import os.path as op
import sys
import math
import matplotlib.pylab as plt
import matplotlib.cm
import matplotlib.colors
from collections import defaultdict
import csv

# Plotting parameters:
NR_CHRS_XTICKS = 30         # number of characters to be printed of the xticks
FIG_SIZE = (15, 7)          # figure size
COLOR_MAP = 'tab20'         # color map. See https://matplotlib.org/users/colormaps.html
#COLOR_MAP = 'Vega10'

import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
import random

cell_types = ["B-cell", "CD4T-cell", "CD8T-cell", "NK-cell", "T-cell",  "granulocyte",  "breast",  "lung", "lung_alveolar", "monocyte", "neuron", "breast_basal", "breast_luminal", "colon", "hepatocyte", "adipocyte", "bladder", "erythrocyte_progenitor", "bone_osteoblast","kidney", "endothelial_cell",  "lung_endothelial", "kidney_endothelial", "vascular_endothelial", "pancreas_endothelial", "prostate", "epiderminal_keratinocyte", "fallopian", "gallbladder", "gastric","head_neck", "heart_cardiomyocyte", "heart_fibroblasts","left_atrium","oligodendrocyte", "ovary", "pancreas_acinar", "pancreas_alpha", "pancreas_beta", "pancreas_delta", "pancreas_duct","skeletal_muscle", "small_intestine", "smooth_muscle", "thyroid", "upper_GI", "uterus_cervix", "colon_fibroblasts", "dermal_fibroblast", "HSAEC", "HCT116", "megakaryocyte", 'pancreas', 'breast']

cell_type_mapping = {
        'Colon-Fibro': 'colon_fibroblasts',
    'Dermal-Fibro': 'dermal_fibroblast',
    "Adipocytes": "adipocyte",
    "Bladder-Ep": "bladder",
    "Blood-B": "B-cell",
    "Blood-Granul": "granulocyte",
    "Blood-Mono+Macro": "monocyte",
    "Blood-NK": "NK-cell",
    "Blood-T": "T-cell",
    "Bone-Osteob": "bone_osteoblast",
    "Breast-Basal-Ep": "breast_basal",
    "Breast-Luminal-Ep": "breast_luminal",
    "Colon-Ep": "colon",
    "Colon-FibroDermal-Fibro": "colon_fibroblasts",
    "Endothel": "endothelial_cell",
    "Epid-Kerat": "epiderminal_keratinocyte",
    "Eryth-prog": "erythrocyte_progenitor",
    "Fallopian-Ep": "fallopian",
    "Gallbladder": "gallbladder",
    "Gastric-Ep": "gastric",
    "Head-Neck-Ep": "head_neck",
    "Heart-Cardio": "heart_cardiomyocyte",
    "Heart-Fibro": "heart_fibroblasts",
    "Kidney-Ep": "kidney",
    "Liver-Hep": "hepatocyte",
    "Lung-Ep-Alveo": "lung_alveolar",
    "Lung-Ep-Bron": "lung",
    "Neuron": "neuron",
    "Oligodend": "oligodendrocyte",
    "Ovary-Ep": "ovary",
    "Pancreas-Acinar": "pancreas_acinar",
    "Pancreas-Alpha": "pancreas_alpha",
    "Pancreas-Beta": "pancreas_beta",
    "Pancreas-Delta": "pancreas_delta",
    "Pancreas-Duct": "pancreas_duct",
    "Megakaryocytes": "megakaryocyte",
    "Prostate-Ep": "prostate",
    "Skeletal-Musc": "skeletal_muscle",
    "Small-Int-Ep": "small_intestine",
    "Smooth-Musc": "smooth_muscle",
    "Thyroid-Ep": "thyroid",
}
color_map = {}
hatch_map = {}
colors = plt.get_cmap(COLOR_MAP).colors
color_map_model = {'nnls': colors[0], 'llse':colors[2], 'mmse': colors[4]}
hatches = [None,'xxx', '///']

for i, cell_type in enumerate(cell_types):
    color_map[cell_type] = colors[i % len(colors)]
    hatch_map[cell_type] = hatches[(i // len(colors)) % len(hatches)]
color_map['other'] = 'black'
hatch_map['other'] = None



def hide_non_blood(df):
    blood_cells = ['tcell', 'erythroblast', 'nkcell', 'bcell', 'progenitor', 'hsc', 'monocyte', 'macrophage', 'eosinophil', 'neutrophil']
    selection = [name not in blood_cells for name in df.index]
    others = df[selection].sum()
    df = df.drop(df.index[tuple([selection])])
    df = df.append(others.rename('other'))
    return df

def aggregate_tissues(df, tissues, name):
    """
    Sum up mixtures of rows in tissues

    :param df: pandas dataframe of mixture proportions
    :tissues: list of tissues to aggregate together
    :return: df with rows aggregated
    """
    tissues = [t for t in tissues if t in df.index]
    # get the rows to aggregate
    # sum them up
    new_row = df.loc[tissues, :].sum().rename(name)
    # drop the rows
    df = df.drop(tissues, axis=0)
    # add the new row and give it the name: name
    df = df.append(new_row)
    return df


def hide_small_tissues(df, keep = [], threshold=0.01):
    """
    tissues with very small contribution are grouped to the 'other' category.
    :return: The DataFrame with the new category ('other'),
             where the low-contribution tissues are set to 0.
    """
    # select tissues with contribution < threshold but not in keep
    criteria = [i for i in df.index if i not in keep]
    tmp = df.loc[criteria, :]
    others = tmp[tmp < threshold].sum()
    # set rows with contribution < threshold to 0
    tmp[tmp < threshold] = 0
    tmp = tmp.loc[(tmp != 0).any(axis=1)]
    tmp = tmp.append(others.rename('other'))
    for row in keep:
        tmp = tmp.append(df.loc[row, :].rename(row))
    return tmp


def gen_bars_colors_hatches(nr_tissues):
    """
    Generate combinations of colors and hatches for the tissues bars
    Every tissue will get a tuple of (color, hatch)
    the last tuple is for the 'other' category, and is always black with no hatch.
    :return: a list of tuples, with length == nr_tissues
    """
    matplotlib.rcParams['hatch.linewidth'] = 0.3
    hatches = [None, 'xxx', '...', 'O', '++'][:nr_tissues // 7]

    nr_colors = int(math.ceil(nr_tissues / len(hatches)) + 1)
    print(nr_colors, hatches)

    # generate bars colors:
    cmap = matplotlib.cm.get_cmap(COLOR_MAP)
    norm = matplotlib.colors.Normalize(vmin=0.0, vmax=float(nr_colors))
    colors = [cmap(norm(k)) for k in range(nr_colors)]
    # for c in range(nr_colors): print(norm(c), cmap(norm(c)), cmap(c))

    def get_i_bar_tuple(i):
        color_ind = i % nr_colors
        hatch_ind = int(i // math.ceil(nr_tissues / len(hatches)))
        return colors[color_ind], hatches[hatch_ind]

    colors_hatches_list = [get_i_bar_tuple(i) for i in range(nr_tissues-1)]
    return colors_hatches_list + [((0, 0, 0, 1), None)]

def g(x):
    if x == 'ct':
        return 0
    else:
        return 1/float(x)


def sort_MLS(x):
    if x == 'ct':
        return 0
    elif 'relapse' in x:
        x_new = "".join(x.split('_')[:2] + ['_'])
        return sum([ord(x0) for x0 in x_new]) + 0.0001
    elif 'BCR' in x:
        return sum([ord(x0) for x0 in x]) + 500
    elif 'add' in x:
        return 1
    else:
        return sum([ord(x0) for x0 in x])

def sort_by_coverage(x):
    """
    sort values by ascending coverage
    """
    try:
        return float(x)
    except ValueError:
        return sum([ord(x0) for x0 in x]) + 32

def plot_res(df, outpath, show=False):

    # df = hide_small_tissues(df, keep = ['lung'])
    # df = hide_non_blood(df)
    nr_tissues, nr_samples = df.shape

    plt.figure(figsize=FIG_SIZE)
    r = [i for i in range(nr_samples)]
    bottom = np.zeros(nr_samples)
    for i in range(nr_tissues):
        plt.bar(r, list(df.iloc[i, :]),
                edgecolor='white',
                width=0.85,
                label=df.index[i],
                bottom=bottom,
                color=color_map[df.index[i]],
                hatch=hatch_map[df.index[i]])
        bottom += np.array(df.iloc[i, :])

    # Custom x axis
    plt.xticks(r, [str(w)[:NR_CHRS_XTICKS] for w in df.columns], rotation='vertical', fontsize=9)
    plt.xlim(-.6, nr_samples - .4)

    # plt.ylabel("Mixture Proportion")
    # plt.xlabel("Percent of HSAEC mixed into PBMC sample")
    # Add a legend and a title
    # make legend below plot
    plt.legend(loc='upper center', bbox_to_anchor=(0.6, -0.1), ncol=7, fontsize=9)
    # plt.title('Deconvolution of in-silico Mixtures of HSAEC and PBMC\n'+ op.basename(outpath))

    # adjust layout, save and show
    plt.tight_layout(rect=[0, 0, .83, 1])
    plt.savefig(outpath)
    if show:
        plt.show()

def plot_scatter(sigma, outpath, cell_types):
    """
    Plot a scatter plot of predicted vs actual proportions

    :param sigma: list of paths to the sigma files
    :param outpath: path to save the plot
    :param cell_types: list of cell types to plot
    :return:
    """
    #initialize a default dict of dicts of lists
    true_to_predicted = defaultdict(lambda: defaultdict(list))
    for s in sigma:

        true_proportion = s.split('/')[-1].split('.')[0].split('_')[0]
        if true_proportion == 'HSAEC':
            continue
        elif true_proportion == 'PBMC':
            true_proportion = 0.0
        else:
            true_proportion = float(true_proportion)/100
        model = s.split('.')[2]
        reader = csv.reader(open(s, 'r'), delimiter='\t')
        agg_proportion = 0.0
        for ct, proportion in reader:
            if ct in cell_types:
                agg_proportion += float(proportion)
        true_to_predicted[model][true_proportion] += [float(agg_proportion)]

    # plot the scatter plot
    plt.figure(figsize=FIG_SIZE)
    for model in true_to_predicted:
        x = [float(k) for k in true_to_predicted[model].keys()]
        y = [np.mean(true_to_predicted[model][k]) for k in x]
        std = [np.std(true_to_predicted[model][k]) for k in x]
        plt.errorbar(x, y, yerr=std, label=model, color=color_map_model[model])
        plt.scatter(x, y, label=model, color=color_map_model[model])
        # draw a line of slope 1
        plt.plot([0, 1], [0, 1], '--', color='grey', alpha=0.5)
        # color the line differently for each model

    # use log base 2 scale for both axes
    plt.loglog(x, y, base=2)
    plt.xlabel('True Proportion')
    plt.ylabel('Predicted Proportion')
    plt.legend(loc='lower center', bbox_to_anchor=(1, 1), ncol=5)
    # Save fig
    plt.tight_layout(rect=[0, 0, .83, 1])
    plt.savefig(outpath)


def plot_mixture_proportions(sigma, outpath, group_lung = False):
    """
    Plot the mixture proportions for each sample in a bar chart.
    plots are labeled by their file name

    :param sigma: list of paths to the sigma files. Each containing one deconvolution result.
    :param outpath: path where to save the figure
    """
    # read all input files:
    dfs = []
    for f in sigma:
        df = pd.read_csv(f, sep='\t', index_col=0)
        dfs.append(df)
    # concatenate all dataframes by column, name columns by input file name:
    df = pd.concat(dfs, axis=1, keys=[op.basename(f) for f in sigma])
    # clean up column names
    new_cols = []
    for c in df.columns:
        name = c[0].split('.')[0]
        if 'seed' in name:
            new_cols.append(str(name).split('_')[0])
        elif 'PBMC' in name:
            new_cols.append(0)
        elif 'HSAEC' in name or 'HCT116' in name:
            new_cols.append(100)
        else:
            new_cols.append(str(name))
    df.columns = new_cols
    # df.columns = [c[0].split('.')[0] for c in df.columns]
    # clean up index names
    df.index = [c.replace(" EPIC", '') for c in df.index]
    if 'Blood-B' in list(df.index):
        df.index = [cell_type_mapping[c] for c in df.index]
    # sort columns by ascending numerical order
    df = df.reindex(sorted(df.columns, key=sort_by_coverage), axis=1)
    # Group lung tissues:
    if group_lung:
        df = aggregate_tissues(df, ['lung', 'lung_alveolar', 'lung_endothelial', 'endothelial_cell'], 'lung')
    plot_res(df, outpath)





import argparse
import os
import sys

script_dir = os.path.dirname(os.path.realpath(__file__))
sys.path.append(script_dir)

from functions import *
from plot import plot_mixture_proportions
BISULFITE_ATLAS = os.path.join(script_dir, '..', '..', 'atlases', '39Bisulfite.tsv')

def main():
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
    description="""
Nanomix: a tool for cell-type deconvolution of mixed methylation calls from Oxford Nanopore sequencing data.
------------------------------------------------------------------------------------------------------------
Version:   v0.1.0
About:     Developed in the simpsonlab.github.io by Jonathan Broadbent.
Code:      https://github.com/simpsonlab/nanomix

This package incorperates 4 different deconvolution models (nnls, llse, llsp, mmse).
Deconvolution results are dependent on the reference atlas used.
We recommend using the atlas from the paper: https://www.biorxiv.org/content/10.1101/2022.01.24.477547v1.full
Which is labelled as '39Bisulfite.tsv' in the 'atlas' directory.

Nanomix sub-commands include:

Deconvolute:    Find the proportion of cell types present in a methylome
Evaluate:       Evaluate the accuracy of a deconvolution model on a given methylome with labelled cell types
Simulate:       Generate a methylome from a given cell-type mixture
Assign:         Assign reads in the methylome to cell-types

All outputs are directed to stdout. Select a sub-command with the [-h] flag for info on hparams
                                     """)
    parser.add_argument('-v', '--version', action='version', version='%(prog)s 0.1.1')
    subparsers = parser.add_subparsers(dest='command', required=True,
                                       help="nanomix functions")
    parser_deconvolute = subparsers.add_parser('deconvolute',formatter_class=argparse.RawDescriptionHelpFormatter,
description="""
Deconvolute a methylome. Return the mixture proportions (sigma) in a tsv format to stdout.

Methylome: A tsv file containing the methylome to be deconvoluted. The file must contain the following columns:
    chr:                The chromosome the read is mapped to
    start:              The start position of the read
    end:                The end position of the read
    total_calls:        The total number of reads mapped to the region
    modified_calls:     The number of modified reads mapped to the region
This file can be created from a BAM file using our associated program, mbtools (https://github.com/jts/mbtools)

Deconvolution Models:
    llse (default):     log-likelihood with sequencing errors. Maximize the likelihood of the methylome, atlas and sigma
                        by assuming modification calls follow a binomial distribution. Good for sequencing data with high error
                        rate and non-uniform coverage distribution. (Oxford Nanopore)
    nnls:               non-negative least squares. Minimize the squared error between the methylome and what we expect for
                        the methylome (given sigma and the atlas). Recommended for fast deconvolution of methylomes with high
                        coverage. (Methylation Arrays)
    mmse:               mixture model with sequencing errors. Also follows a binomial distribution, but softly assigns fragments
                        to cell-types. Optimization uses expectation maximization (slower than above). Recommended for high resolution
                        deconvolution (many cell types) and an atlas with large regions of grouped CpGs.
    llsp:               log-likelihood with sequencing perfect. Same as llse, without error modelling. Useful for differentiating the
                        effect of sequencing errors on deconvolution loss and accuracy.
""")
    parser_deconvolute.add_argument('methylome', help='Path to methylome tsv file with columns: {chr, start, end, total_calls, modified_calls}')
    parser_deconvolute.add_argument('-a', '--atlas', required=True, type=str, default=BISULFITE_ATLAS, help='Path to reference atlas')
    parser_deconvolute.add_argument('-p01', default=0.05, help='Sequencing miscall rate')
    parser_deconvolute.add_argument('-p11', default=0.95, help='Sequencing correct call rate')
    parser_deconvolute.add_argument('-m', '--model', default='llse', type=str, help='Deconvolution model options: [nnls, llse, llsp, mmse]')
    parser_deconvolute.add_argument('-@', '--threads', default=1, type=int, help='Number of threads to use')
    parser_deconvolute.add_argument('-nt', '--n_trials', default=3, type=int, help='Number of trials to run')
    parser_deconvolute.add_argument('--nnls_init' , action='store_true', help='Initialize nnls with llse')
    parser_deconvolute.add_argument('--concentration', type=float, help='Concentration parameter for Dirichlet prior')
    parser_deconvolute.add_argument('-i', '--sigma_init', default='null', type=str, help='Initalize sigma with a tsv file')
    parser_deconvolute.add_argument('-n', '--max_iter', default=200, type=int, help='Maximum number of iterations for the EM optimization (mmse only)')
    parser_deconvolute.add_argument('-p', '--min_proportion', default=0.01, type=float, help='Minimum proportion of a cell type to be considered (mmse only)')
    parser_deconvolute.add_argument('-t', '--stop_thresh', default=1e-5, type=float, help='Stop EM iterations when percent change in log-likelihood falls below this value (mmse only)')
    parser_deconvolute.set_defaults(func=deconvolute)

    parser_evaluate = subparsers.add_parser('evaluate', formatter_class=argparse.RawDescriptionHelpFormatter, description="""
Evaluate the loss and accuracy of cell-type deconvolution on a given methylome with known cell types.

Metrics outputed to stdout:
    Deconvolution loss:         Euclidean distance between the computed and true mixture proportion.
    Accuracy at confidence <t>: Proportion of correctly assigned reads with confience >= t.


Methylome: A tsv file containing the methylome to be deconvoluted. The file must contain the following columns:
    chr:                The chromosome the read is mapped to
    start:              The start position of the read
    end:                The end position of the read
    total_calls:        The total number of reads mapped to the region
    modified_calls:     The number of modified reads mapped to the region
    cell_type:          String identifier of cell_type that matches that in the atlas header
The simulate function can be used to create this file.
""")
    parser_evaluate.add_argument('methylome', help='Path to methylome tsv file with columns: {read_name, chr, start_position, end_position, total_calls, modified_calls, cell_type}')
    parser_evaluate.add_argument('-a', '--atlas', required=True, type=str, default=BISULFITE_ATLAS, help='Path to reference atlas')
    parser_evaluate.add_argument('-s', '--sigma', type=str, help='Path to sigma tsv file')
    parser_evaluate.add_argument('-p01', default=0.05, type=float, help='Sequencing miscall rate')
    parser_evaluate.add_argument('-p11', default=0.95, type=float, help='Sequencing correct call rate')
    parser_evaluate.set_defaults(func=evaluate)

    parser_simulate = subparsers.add_parser('simulate', formatter_class=argparse.RawDescriptionHelpFormatter, description="""
Simulate a methylome from a given cell-type mixture (sigma).

Sigma: A tsv file containing the cell-type mixture proportions. The file must contain the following columns:
    cell_type:          String identifier of cell_type that matches that in the atlas header
    proportion:         float between 0 and 1
""")
    parser_simulate.add_argument('sigma', type=str, help='Path to sigma tsv file')
    parser_simulate.add_argument('-a', '--atlas', required=True, type=str, default=BISULFITE_ATLAS, help='Path to reference atlas')
    parser_simulate.add_argument('-p01', default=0.05, type=float, help='Sequencing miscall rate')
    parser_simulate.add_argument('-p11', default=0.95, type=float, help='Sequencing correct call rate')
    parser_simulate.add_argument('-c', '--coverage', default=1, type=float, help='Sequencing coverage')
    parser_simulate.add_argument('-r', '--region_size', default=5, type=int, help='Number of CpGs in each region')
    parser_simulate.set_defaults(func=simulate)

    parser_assign = subparsers.add_parser('assign', formatter_class=argparse.RawDescriptionHelpFormatter, description="""
Assign cell types to a methylome using a given atlas and pre-deconvoluted sigma.
Output (to stdout) is of the same format as the methylome with an appended column indicating the cell_type.

Methylome:  A tsv file containing the methylome to be deconvoluted. The file must contain the following columns:
    chr:                The chromosome the read is mapped to
    start:              The start position of the read
    end:                The end position of the read
    total_calls:        The total number of reads mapped to the region
    modified_calls:     The number of modified reads mapped to the region
    cell_type:          String identifier of cell_type that matches that in the atlas header
The simulate function can be used to create this file.
""")
    parser_assign.add_argument('methylome', help='Path to methylome tsv file with columns: {chr, start, end, total_calls, modified_calls}')
    parser_assign.add_argument('-a', '--atlas', required=True, type=str, default=BISULFITE_ATLAS, help='Path to reference atlas')
    parser_assign.add_argument('-p01', default=0.05, type=float, help='Sequencing miscall rate')
    parser_assign.add_argument('-p11', default=0.95, type=float, help='Sequencing correct call rate')
    parser_assign.add_argument('-s', '--sigma', required=True, type=str, help='Path to sigma tsv file')
    parser_assign.set_defaults(func=assign_fragments)

    parser_plot = subparsers.add_parser('plot', formatter_class=argparse.RawDescriptionHelpFormatter, description="""
Plot the results of a deconvolution into a stacked bar chart.

Provide a list of sigma files to combine plots into one chart
Sigma: A tsv file containing the cell-type mixture proportions. The file must contain the following columns:
    cell_type:          String identifier of cell_type that matches that in the atlas header
    proportion:         float between 0 and 1
""")
    parser_plot.add_argument('sigma', nargs='+', type=str, help='Path to sigma tsv files')
    parser_plot.add_argument('-o', '--outpath', type=str, help='Name of the plot')
    parser_plot.add_argument('-l', '--group_lung', action='store_true', help='Group lung cell types into one category')
    parser_plot.add_argument('-c', '--chart', type=str, default='bar', help='Chart type: [bar, scatter]')
    parser_plot.add_argument('-ct', '--cell_types', nargs='+', type=str, help='List of cell types to plot')
    parser_plot.set_defaults(func=plot)

    parser_ll = subparsers.add_parser('ll', formatter_class=argparse.RawDescriptionHelpFormatter, description="""
Compute the log-likelihood of a given methylome given a cell-type mixture (sigma).
""")
    parser_ll.add_argument('methylome', help='Path to methylome tsv file with columns: {chr, start, end, total_calls, modified_calls}')
    parser_ll.add_argument('-a', '--atlas', required=True, type=str, default=BISULFITE_ATLAS, help='Path to reference atlas')
    parser_ll.add_argument('-m', '--model', default='llse', type=str, help='Deconvolution model options: [llse, llsp, mmse]')
    parser_ll.add_argument('-p01', default=0.05, type=float, help='Sequencing miscall rate')
    parser_ll.add_argument('-p11', default=0.95, type=float, help='Sequencing correct call rate')
    parser_ll.add_argument('-s', '--sigma', required=True, type=str, help='Path to sigma tsv file')
    parser_ll.set_defaults(func=log_likelihood)



    args = parser.parse_args()
    arg_dict = {k : v for k, v in vars(args).items() if k not in ['func','command']}
    args.func(**arg_dict)

if __name__ == "__main__":
    main()
